"""
------------------------------------------------
FALL 2008 :: COMP522 FINAL PROJECT
Agent-based Modelling with Statecharts
Studying the behaviour of spatially distributed systems using collections of interacting Statecharts
See assosciated website:
http://msdl.cs.mcgill.ca/people/hv/teaching/MS/projects/material/Nabeelah.Ali/

Nabeelah Ali
nabeelah.ali@mail.mcgill.ca

December 2008
------------------------------------------------
"""

# Model compiled from AToM3.
# For more information on AToM3, go here -> http://atom3.cs.mcgill.ca/
from simpleDisease_MDL_Compiled import simpleDisease_MDL

from random import *
from time import *

# The numpy module is needed for array-handling. For more information, go here -> http://numpy.scipy.org/
from numpy import *

# For the GUI.
import pygame, sys,os
from pygame.locals import *
import wx

"""
The landscape for my cellular automaton is a square grid. The grid size can
be specified by altering the value below. The grid is represented as an array,
with values meaning the following things:
    * 0 = empty grid spot
    * 1 = healty occupant
    * 2 = diseased occupant
    * 3 = dead occupant
"""
grid_size = 30
people = 120
landscape = zeros((grid_size, grid_size))


class simpleDisease:

    def __init__(self):
        self.statechart = simpleDisease_MDL()
        self.statechart.initModel()
        self.controller = simpleDisease_Controller(self.statechart)
        self.statechart.event("start", self.controller)
        #print "Initializing a new person.", self.statechart.get_current_state()


class simpleDisease_Controller:
    def __init__(self, statechart):
        self.statechart = None
        self.state = 1

        self.statechart = statechart
        
        # Set a random location for this new agent.
        found = False
        # Make sure the grid spot is empty.
        while (found == False):
            self.location = (randint(0,grid_size - 1),randint(0,grid_size - 1))
            if landscape[self.location] == 0:
                found = True
        # Mark the grid.
        landscape[self.location] = self.state
        self.numBuried = 0

    def infection(self):
        self.statechart.event('infection')

    def makeInfected(self):
        self.state = 2
        landscape[self.location] = self.state
        updateGUI(None,self.location,self.state)

    def makeHealthy(self):
        self.state = 1
        landscape[self.location] = self.state
        updateGUI(None,self.location,self.state)

    def bury(self):
        landscape[self.location] = 0
        updateGUI(None,self.location,0)

    def die(self):
        self.state = 3
        landscape[self.location] = self.state
        updateGUI(None,self.location,self.state)

    def move(self):
        # print "I am moving now!"

        oldLocation = self.location
        empty = False

        # Move in a random direction while ensuring it is not occupied.
        # What if all grid spots around a particular agent are occupied?
        # The next interrupt from a time-advance will stop the processing
        # and so the agent will stay in the same place.
        
        """
        Movement directions.
         1 | 2 | 3
         __|___|__
           |   |
         4 | O | 5
         __|___|__
           |   |
         6 | 7 | 8
        """

        while (empty == False):
            # Select a direction in which to randomly move.
            newLoc = randint(1,8)

            if (newLoc == 1):
                # Move to top-left grid squre.
                newLocation = (self.location[0] - 1, self.location[1] - 1)
            elif (newLoc == 2):
                # Move to top grid square.
                newLocation = (self.location[0], self.location[1] - 1)
            elif (newLoc == 3):
                # Move to top-right grid square.
                newLocation = (self.location[0] + 1, self.location[1] - 1)
            elif (newLoc == 4):
                # Move to left grid square.
                newLocation = (self.location[0] - 1, self.location[1])
            elif (newLoc == 5):
                # Move to right grid square.
                newLocation = (self.location[0] + 1, self.location[1])
            elif (newLoc == 6):
                # Move to bottom-left grid square.
                newLocation = (self.location[0] - 1, self.location[1] + 1)
            elif (newLoc == 7):
                # Move to bottom grid square.
                newLocation = (self.location[0], self.location[1] + 1)
            elif (newLoc == 8):
                # Move to bottom-right grid square.
                newLocation = (self.location[0] + 1, self.location[1] + 1)

            # Checks to make sure location does not exceed grid size. The grid
            # 'wraps around' like a globe - there are no edges so agents that disappear
            # off one end should reappear on the other side.
            
            if (newLocation[0] == grid_size):
                newLocation = (0, newLocation[1])
            elif (newLocation[0] < 0):
                newLocation = (grid_size - 1, newLocation[1])
                
            if (newLocation[1] == grid_size):
                newLocation = (newLocation[0], 0)
            elif (newLocation[1] < 0):
                newLocation = (newLocation[0], grid_size - 1)

            # Error checking.
            # print "The new location is", newLocation
            
            if landscape[newLocation] != 0:
                #print "Attempting to move to a location that is not empty."
                empty = False
            else:
                empty = True
                self.location = newLocation

        # Erase old location and mark new location.      
        landscape[oldLocation] = 0
        landscape[newLocation] = self.state
        # print landscape
        updateGUI(oldLocation, newLocation, self.state)

        # Now check out who's in the vicinity, and store the states of occupants in an array.
        stateArray = zeros(8)
        x = self.location[0]
        y = self.location[1]
        minus1x = x - 1
        minus1y = y - 1
        add1x = x + 1
        add1y = y + 1
        if ((x + 1) == grid_size): add1x = 0
        if ((y + 1) == grid_size): add1y = 0
        if ((x - 1) < 0): minus1x = grid_size - 1
        if ((y - 1) < 0): minus1y  = grid_size - 1

        stateArray[0] = landscape[(minus1x, minus1y)]
        stateArray[1] = landscape[(x, minus1y)]
        stateArray[2] = landscape[(add1x, minus1y)]
        stateArray[3] = landscape[(minus1x, y)]
        stateArray[4] = landscape[(add1x, y)]
        stateArray[5] = landscape[(minus1x, add1y)]
        stateArray[6] = landscape[(x, add1y)]
        stateArray[7] = landscape[(add1x, add1y)]

        # Iterate through array.
        # If next to an infected person, trigger the 'infection' event.
        # If next to a dead person, trigger to 'death' event.
        
        for i in stateArray:
            if i == 1:
                #print "Next to a person."
                pass
            elif i == 2:
                #print "Next to an infected person!"
                self.statechart.event("infection")
            elif i == 3:
                #print "Bring out the funeral flowers."
                self.statechart.event("death")
                
        #print stateArray
        

def updateGUI(oldLoc, newLoc, state):
    
    if (state == 1):
        icon = healthyIcon
    elif (state == 2):
        icon = infectedIcon
    elif (state == 3):
        icon = deadIcon
    elif (state == 0):
        icon = emptyIcon

    
    screen.blit(icon, (newLoc[0] * iconSize, newLoc[1] * iconSize))
    rectList1 = (newLoc[0] * iconSize, newLoc[1] * iconSize, (newLoc[0] + iconSize) * iconSize, (newLoc[1] + iconSize) * iconSize)

    numInfected = 0
    numDead = 0
    numHealthy = 0

    for agentRow in landscape:
        for agent in agentRow:
            if agent == 1:
                numHealthy += 1
            elif agent == 2:
                numInfected += 1
            elif agent == 3:
                numDead += 1
    numBuried = people - (numHealthy + numInfected + numDead)
    
    #pygame.display.update(rectList)
    displayText = "  Healthy: " + str(numHealthy) + "  Infected: " + str(numInfected) + "  Dead: " + str(numDead) + "  Buried: " + str(numBuried)
    text = font.render(displayText, 1, (10, 10, 10))
    textpos = (0,windowSize + padding)
    # Clear the screen.
    screen.fill(0xFFFFFF, (0, windowSize + padding, windowSize, windowSize + textSize + padding))
    screen.blit(text, textpos)
    rectList3 = (0, windowSize + padding, windowSize, windowSize + textSize + padding)

    if (oldLoc != None):
        screen.blit(emptyIcon, (oldLoc[0] * iconSize, oldLoc[1] * iconSize))
        rectList2 = (oldLoc[0] * iconSize, oldLoc[1] * iconSize, (oldLoc[0] + iconSize) * iconSize,  (oldLoc[1] + iconSize) * iconSize)

        pygame.display.update((rectList1, rectList2, rectList3))
    else:
        pygame.display.update((rectList1, rectList3))

    """
    events = pygame.event.get()
    for e in events:
        if (e.type == KEYDOWN):
            if( e.key == K_ESCAPE ):
                done = True
                break
    """
    

    #

    #pygame.display.update()
    
    
if __name__=="__main__":
 

  app = wx.App()

  frame = wx.Frame(None, -1, 'simple.py')
  frame.Show()

  app.MainLoop()

  
    
  numBuried = 0
  # Set the display.

  # Icon size.
  iconSize = 16
  windowSize = grid_size * iconSize
  textSize = 30
  padding = 10
  
  pygame.init()
  window = pygame.display.set_mode((windowSize, windowSize + textSize + (padding * 2))) 
  pygame.display.set_caption('Disease Spreading')
  pygame.mouse.set_visible(0)
  screen = pygame.display.get_surface()
  screen.fill((255, 255, 255))

  # Icons used here are under a Creative Commons license.
  # Creator: http://www.famfamfam.com/lab/icons/silk/
  healthyIconFile = os.path.join("images","healthy.png")
  infectedIconFile = os.path.join("images","infected.png")
  emptyIconFile = os.path.join("images","empty.png")
  deadIconFile = os.path.join("images","dead.png")
  
  healthyIcon = pygame.image.load(healthyIconFile).convert()
  infectedIcon = pygame.image.load(infectedIconFile).convert()
  emptyIcon = pygame.image.load(emptyIconFile).convert()
  deadIcon = pygame.image.load(deadIconFile).convert()

  for xC in xrange(0,grid_size):
      for yC in xrange(0,grid_size):
          screen.blit(emptyIcon, (xC * iconSize, yC * iconSize))
          #sleep(0.2)
    
  font = pygame.font.Font(None, textSize)


  pygame.display.update() 

  for j in xrange(0,people):
      person = simpleDisease()
      inf = uniform(0,2)
      if inf < 0.5:
          person.controller.statechart.event("infection")

  #person3.controller.statechart.event("infection")
  
  while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            raise SystemExit
  

"""
def input(events): 
   for event in events: 
          if event.type == QUIT: 
                 sys.exit(0) 
          else: 
                 print event 
 
while True: 
   input(pygame.event.get()) 
"""
